#ifndef DTSS_H_
#define DTSS_H_


class DTSS{
public:
	float maxValue;
	int set_cardinality;
	float node_prob;
	char* dataset_file_name;
	int TO_dimension;
	int PO_dimension;
	int data_cardinality;
	char dataset_type;

	// derived
	string scaledDataFileName; // filename of the scaled dataset
	string scaledDataFileNamePath; // path to file of the scaled dataset
	string outDirectory; // directory to output method related files

	// constructed
	Graph **G; // DAGs for PO dimensions
	multimap<int, string> dictionary; // the dictionary information
	map<string, int> groupSizes; // <group_id, num of points in stratum>
	ofstream progr; // filestream for progressiveness
	DTSS_RTree **RT;
	map<int, string> treeNames; // <tree ID, tree name>

	// results
	vector<float> result;

	// statistics
//	ProcessInfo pi;
	int numResults;
	int accesses;
	long query_time;
	long total_time;
	int numTotalLeafNodes;
	int numTotalInternalNodes;


	// METHODS

	// con- (de-) structors
	DTSS(float maxValue, int set_cardinality, float node_prob, char* dataset_file_name, int TO_dimension, int PO_dimension, int data_cardinality, char dataset_type);
	~DTSS();



	// dominance checks
	bool PO_dominate(float* _pt1, float* _pt2);


	void initialize();
	void execute();
	void skyline();
	void assignStringToNum(float* n, string s, int dim);
	void insertIntoFile(ofstream *fileout, float* point);
	void flushIntoFile(string group, vector<float *>& points);
	void createSubDatasets();
	void createSubDatasetsOLD();
	string determineGroup(float *PO_point);
	int computeRectDist(string group);
	void createRTrees();

	void printResults();
	string getStatistics();
};

#endif /*DTSS_H_*/
