#include "../common/pch.h"

bool DTSS_RTree::sky_dom_iter(float* _pt){ // _pt contains TO_dimension + 2*PO_dimension floats (a TO point and a PO MBR)
	int TO_dim = dtss->TO_dimension;
	int PO_dim = dtss->PO_dimension;
	int i, j;
	bool ret = false;	
	vector<float>::iterator iter;

	iter = dtss->result.begin();

	// Extract TO_point
	float *TO_point = new float[TO_dim];
	for(i=0;i<TO_dim;i++)
		TO_point[i] = _pt[i];
	
	while (iter!=dtss->result.end()){
		// Extract a result point
		float *s_pt = new float[TO_dim+PO_dim];
		for (int j=0; j<(TO_dim+PO_dim); j++){
			s_pt[j] = *iter; 
			iter++;
		}

		// Check dominance on the total dimensions
		if (!TO_dominate(TO_dim, s_pt, TO_point) || !dtss->PO_dominate(s_pt+TO_dim, _pt+TO_dim)){
			delete[] s_pt;
			continue;
		}
		
		delete[] s_pt;
		ret = true;
		break;
	}

	delete[] TO_point;
	return ret;
}

void DTSS_RTree::skyline(XxkHeap *_hp, float *PO_values){
	int TO_dim = dtss->TO_dimension;
	int PO_dim = dtss->PO_dimension;
	int i, j;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greather than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j++)
		he->bounces[j] = 0.0;
	_hp->push( he );

	float *b = new float[TO_dim+2*PO_dim];
	float *bounces = new float[TO_dim+2*PO_dim];

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;

		for (i = 0; i < TO_dim; i ++){
			bounces[i] = he->bounces[2*i];
		}
		for(i=0;i<PO_dim;i++){
			bounces[TO_dim+2*i] = PO_values[i];
			bounces[TO_dim+2*i+1] = PO_values[i];
		}
		
		if (level==0){
			if (!sky_dom_iter(bounces)){
				for(j=0;j<TO_dim;j++){
					//cout<<bounces[j]<<" ";
					dtss->result.push_back(bounces[j]);
				}
				for(j=0;j<PO_dim;j++){
					//cout<<bounces[TO_dim+2*j]<<" ";
					dtss->result.push_back(bounces[TO_dim+2*j]);
				}
				//cout<<endl;
			}
		}
		else{
			if (!sky_dom_iter(bounces)){
				RTNode *child = new RTNode(this, son);
				dtss->accesses++;
				
				for (i = 0; i < child->num_entries; i ++){
					for (j = 0; j < TO_dim; j ++) {
						b[j] = child->entries[i].bounces[2 * j];
					}
					for(j=0;j<PO_dim;j++){
						b[TO_dim+2*j] = PO_values[j];
						b[TO_dim+2*j+1] = PO_values[j];
					}
					if (!sky_dom_iter(b)){
						HeapEntry *tmp_he = new HeapEntry();
						tmp_he->init_HeapEntry( dimension );
						tmp_he->son1 = child->entries[i].son;
						tmp_he->level = child->level;
						tmp_he->key = 0;
						
						for (j = 0; j < dimension; j ++)
							tmp_he->key += b[j]; // Manhattan distance

						memcpy( tmp_he->bounces, child->entries[i].bounces, 2 * dimension * sizeof( float ) );
						_hp->push( tmp_he );
					}
				}
				delete child;
			}
		}
		delete he;
	}
	delete [] bounces;
	delete [] b;

	return;
}

