#ifndef SDC_H_
#define SDC_H_


class SDC{
public:
	float maxValue;
	int set_cardinality;
	float node_prob;
	char* dataset_file_name;
	int TO_dimension;
	int PO_dimension;
	int data_cardinality;
	char dataset_type;

	// derived
	string scaledDataFileName; // filename of the scaled dataset
	string scaledDataFileNamePath; // path to file of the scaled dataset
	string outDirectory; // directory to output method related files

	// constructed
	Graph **G; // DAGs for PO dimensions
	vector<int> dictionary; // the dictionary information
	map<int, int> strataSizes; // <stratum_id, num of points in stratum>
	ofstream progr; // filestream for progressiveness
	SDC_RTree **RT;

	// results
	list<float> global_result;
	list<float> local_result;

	// statistics
	Timing *totim;
	int numResults;
	int false_hits;
	int accesses;
	long query_time;
	long total_time;
	int numTotalLeafNodes;
	int numTotalInternalNodes;


	// METHODS

	// con- (de-) structors
	SDC(float maxValue, int set_cardinality, float node_prob, char* dataset_file_name, int TO_dimension, int PO_dimension, int data_cardinality, char dataset_type);
	~SDC();


	// dominance checks
	bool dominate(float *_pt1, float *_pt2);
	bool m_dominate(int _dim, float *_pt1, float *_pt2);


	void initialize();
	void execute();
	void insertIntoFile(ofstream *fileout, float* point);
	float *convertPoint(float* point);
	void createSubDatasets();
	int determineStratum(float *PO_point);
	void createRTrees();
	void skyline();
	void mergeResults_iter_simple();
	void mergeResults_iter();

	void printResults();
	string getStatistics();
};

#endif /*SDC_H_*/
