#ifndef TSS_H_
#define TSS_H_


class TSS{
public:
	float maxValue;
	int set_cardinality;
	float node_prob;
	char* dataset_file_name;
	int TO_dimension;
	int PO_dimension;
	int data_cardinality;
	char dataset_type;

	// derived
	string scaledDataFileName; // filename of the scaled dataset
	string scaledDataFileNamePath; // path to file of the scaled dataset
	string outDirectory; // directory to output method related files

	// constructed
	Graph **G; // DAGs for PO dimensions
	ofstream progr; // filestream for progressiveness
	TSS_RTree *RT;

	// results
	vector<float> result;

	// statistics
	Timing *totim;
	int numResults;
	int accesses;
	long query_time;
	long total_time;
	int numLeafNodes;
	int numInternalNodes;


	// METHODS

	// con- (de-) structors
	TSS(float maxValue, int set_cardinality, float node_prob, char* dataset_file_name, int TO_dimension, int PO_dimension, int data_cardinality, char dataset_type);
	~TSS();


	// dominance checks
	bool PO_dominateOLD(float* _pt1, float* _pt2);
	bool PO_dominate(float* _pt1, float* _pt2);


	void initialize();
	void execute();
	void skyline();

	string getStatistics();
	void printResults();
};

#endif /*TSS_H_*/
