#include "pch.h"
#include <sys/stat.h>

string getDirectory(char *dataset_file_name){
	string dataFileName(dataset_file_name);
	int lastDir = dataFileName.find_last_of("/");
	if (lastDir == string::npos){
		lastDir = 0;
	}
	else {
		lastDir++;
	}
	return dataFileName.substr(0, lastDir);
}

string getFileName(char *dataset_file_name){
	string dataFileName(dataset_file_name);
	int lastDir = dataFileName.find_last_of("/");
	if (lastDir == string::npos){
		lastDir = 0;
	}
	else {
		lastDir++;
	}
	return dataFileName.substr(lastDir);
}


string getDirSuffix(string fileName){
	ostringstream dirSuffixStr;
	dirSuffixStr << fileName << ".dir";
//	dirSuffixStr << ""; // same directory as data set
	return dirSuffixStr.str();
}


string getScaledFileName(char *dataset_file_name, int set_cardinality, float node_prob, int TO_dimension, int PO_dimension){
	string directory, fileName, dirSuffix;
	directory = getDirectory(dataset_file_name);
	fileName = getFileName(dataset_file_name);
	dirSuffix = getDirSuffix(fileName);
	
	// create output Directory if it does not exist
	string outDirectory = directory + dirSuffix;
//	CreateDirectory(outDirectory.c_str(), NULL);
	create_directory(outDirectory.c_str());
	
	ostringstream scaledDataset;
//	scaledDataset << directory << dirSuffix << "/";
	scaledDataset << fileName << "_po" << PO_dimension;
	scaledDataset << "_sc" << set_cardinality << "_p" << node_prob;
	return scaledDataset.str();
}

string getScaledFileNamePath(char *dataset_file_name, int set_cardinality, float node_prob, int TO_dimension, int PO_dimension){
	string directory, fileName, dirSuffix;
	directory = getDirectory(dataset_file_name);
	fileName = getFileName(dataset_file_name);
	dirSuffix = getDirSuffix(fileName);
	
	// create output Directory if it does not exist
	string outDirectory = directory + dirSuffix;
//	CreateDirectory(outDirectory.c_str(), NULL);
	create_directory(outDirectory.c_str());
	
	ostringstream scaledDataset;
	scaledDataset << directory << dirSuffix << "/";
	scaledDataset << fileName << "_po" << PO_dimension;
	scaledDataset << "_sc" << set_cardinality << "_p" << node_prob;
	return scaledDataset.str();
	
}


string getOutDirectory(char *dataset_file_name, string method, string scaledFileName){
	string directory, fileName, dirSuffix;
	directory = getDirectory(dataset_file_name);
	fileName = getFileName(dataset_file_name);
	dirSuffix = getDirSuffix(fileName);
	
	ostringstream outDirStr;
	outDirStr << directory << dirSuffix << "/" << scaledFileName << ".dir";
//	CreateDirectory(outDirStr.str().c_str(), NULL);
	create_directory(outDirStr.str().c_str());
	outDirStr << "/" << method;
//	CreateDirectory(outDirStr.str().c_str(), NULL);
	create_directory(outDirStr.str().c_str());
	outDirStr << "/";
	
	return outDirStr.str();
}


string initiateScaledFile(char *dataset_file_name, float maxValue, int set_cardinality, float node_prob, int TO_dimension, int PO_dimension, int* graphSizes){
	string directory, fileName, dirSuffix;
	directory = getDirectory(dataset_file_name);
	fileName = getFileName(dataset_file_name);
	dirSuffix = getDirSuffix(fileName);
	
	string scaledDataFileName = getScaledFileName(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	scaledDataFileName = directory + dirSuffix + "/" + scaledDataFileName;
	
	cerr << "scaled file is " << scaledDataFileName << endl;
	
	// create scaled file only if not found!
	ifstream inp;
	inp.open(scaledDataFileName.c_str(), ifstream::in);
	inp.close();
	if(inp.fail()){
		inp.clear(ios::failbit);
		//cerr << "gonna create scaled file" << endl;
		// create scaledFile
		createScaledInputFile(dataset_file_name, scaledDataFileName, TO_dimension, PO_dimension, graphSizes, maxValue);
		cerr << "scaled file created" << endl;
	}
	else{
		cerr << "scaled file exists" << endl;
	}
	
	return scaledDataFileName;
}


void createScaledInputFile(char *dataset_file_name, string scaledDataFileName, int TO_dimension, int PO_dimension, int *graphSizes, float maxValue){
	ifstream fin;
	fin.open(dataset_file_name);
	
	ofstream fout;
	fout.open(scaledDataFileName.c_str(), ios::out);
	
	bool exitloop = false;
	while(!fin.eof()){
		float dummy;
		for (int i=0; i<TO_dimension; i++){
			fin >> dummy;
			if(fin.eof()){
				exitloop = true;
				break;
			}
			fout << dummy << " ";
		}
		if (exitloop) break;
		for (int i=0; i<PO_dimension; i++){
			fin >> dummy;
			fout << scale(dummy, graphSizes[i], maxValue) << " ";
		}
		fout << endl;
	}
	
	fin.close();
	fout.close();
}

int scale(float value, int graphSize, float maxValue){
	int index = (int) ( (value*graphSize)/(maxValue+FLOATZERO) ); // scale
	if (index==graphSize) {
		index = graphSize - 1;
	}
	return index;
}

bool TO_dominate(int _dim, float *_pt1, float *_pt2){
	bool ret = true;
	for (int i = 0; i < _dim; i ++){
		if (_pt1[i] > _pt2[i] + FLOATZERO){
			ret = false;
			break;
		}
	}
	return ret;
}
