#include "pch.h"





int main(int argc, char** argv){

	// hard-coded values DEPRECATED
	float maxValue = 10000;
	int set_cardinality = 5;
	float node_prob = 0.6;
	char* dataset_file_name= "C:\\Users\\Stavros\\Documents\\TSS\\iD3T1M10000C500000";
	int TO_dimension = 2;
	int PO_dimension = 1;
	int data_cardinality = 500000;
	char dataset_type = 'i';
	string valueToPrint = "";
	bool runSDC = true, runTSS = true, runDTSS = true;

	// values read from stdin
	if (argc>0){
		runSDC = false;
		runTSS = false;
		runDTSS = false;
		if (argc<10){
			cerr << "Usage: " << argv[0] << " [input_file] [maxValue] [set_cardinality] [node_prob] [TO_dimension] [PO_dimension] [data_cardinality] [dataset_type] [valueToPrint] [methods to run (TSS, SDC, DTSS)]" << endl;
			exit(0);
		}
		dataset_file_name = argv[1];
		maxValue = atof(argv[2]);
		set_cardinality = atoi(argv[3]);
		node_prob = atof(argv[4]);
		TO_dimension = atoi(argv[5]);
		PO_dimension = atoi(argv[6]);
		data_cardinality = atoi(argv[7]);
		dataset_type = argv[8][0];
		valueToPrint = argv[9];
		for (int i=9; i<argc; i++){
			string argument(argv[i]);
			if (argument.compare("SDC")==0){
				runSDC = true;
			}
			if (argument.compare("TSS")==0){
				runTSS = true;
			}
			if (argument.compare("DTSS")==0){
				runDTSS = true;
			}
		}
	}


	// result files
	string resultsDir = "results"; // hard-coded
	string falseHitsFilename, timeFilename, totimeFilename, iosFilename, sizeFilename, nLeavesFilename, nInternalFilename;
	falseHitsFilename = resultsDir + "/falseHits.txt";
	timeFilename = resultsDir + "/time.txt";
	totimeFilename = resultsDir + "/totime.txt";
	iosFilename = resultsDir + "/IOs.txt";
	sizeFilename = resultsDir + "/skylineSize.txt";
	nLeavesFilename = resultsDir + "/nLeaves.txt";
	nInternalFilename = resultsDir + "/nInternal.txt";

	ofstream falseHitsOut(falseHitsFilename.c_str(), ios::app);
	ofstream timeOut(timeFilename.c_str(), ios::app);
	ofstream totimeOut(totimeFilename.c_str(), ios::app);
	ofstream iosOut(iosFilename.c_str(), ios::app);
	ofstream sizeOut(sizeFilename.c_str(), ios::app);
	ofstream nLeavesOut(nLeavesFilename.c_str(), ios::app);
	ofstream nInternalOut(nInternalFilename.c_str(), ios::app);

	falseHitsOut << valueToPrint << "\t";
	timeOut << valueToPrint << "\t";
	totimeOut << valueToPrint << "\t";
	iosOut << valueToPrint << "\t";
	sizeOut << valueToPrint << "\t";
	nLeavesOut << valueToPrint << "\t";
	nInternalOut << valueToPrint << "\t";


	// SDC
	if (runSDC){
		SDC *sdc = new SDC(maxValue, set_cardinality, node_prob, dataset_file_name, TO_dimension, PO_dimension, data_cardinality, dataset_type);
		cerr << "Executing SDC..." << endl;
		sdc->execute();
//		cerr << "Outputing statistics..." << endl;
		cerr << sdc->getStatistics();
		falseHitsOut << sdc->false_hits << "\t";
		timeOut << sdc->query_time << "\t";
		totimeOut << sdc->total_time << "\t";
		iosOut << sdc->accesses << "\t";
		sizeOut << sdc->numResults << "\t";
		nLeavesOut << sdc->numTotalLeafNodes << "\t";
		nInternalOut << sdc->numTotalInternalNodes << "\t";
		falseHitsOut.flush();
		timeOut.flush();
		totimeOut.flush();
		iosOut.flush();
		sizeOut.flush();
		nLeavesOut.flush();
		nInternalOut.flush();
	}

	// TSS
	if (runTSS){
		TSS *tss = new TSS(maxValue, set_cardinality, node_prob, dataset_file_name, TO_dimension, PO_dimension, data_cardinality, dataset_type);
		cerr << "Executing TSS..." << endl;
		tss->execute();
//		cerr << "Outputing statistics..." << endl;
		cerr << tss->getStatistics();
		timeOut << tss->query_time << "\t";
		totimeOut << tss->total_time << "\t";
		iosOut << tss->accesses << "\t";
		sizeOut << tss->numResults << "\t";
		nLeavesOut << tss->numLeafNodes << "\t";
		nInternalOut << tss->numInternalNodes << "\t";
		falseHitsOut.flush();
		timeOut.flush();
		totimeOut.flush();
		iosOut.flush();
		sizeOut.flush();
		nLeavesOut.flush();
		nInternalOut.flush();
	}

	// DTSS
	if (runDTSS){
		DTSS *dtss = new DTSS(maxValue, set_cardinality, node_prob, dataset_file_name, TO_dimension, PO_dimension, data_cardinality, dataset_type);
		cerr << "Executing DTSS..." << endl;
		dtss->execute();
//		cerr << "Outputing statistics..." << endl;
		cerr << dtss->getStatistics();
		timeOut << dtss->query_time << "\t";
		totimeOut << dtss->total_time << "\t";
		iosOut << dtss->accesses << "\t";
		sizeOut << dtss->numResults << "\t";
		nLeavesOut << dtss->numTotalLeafNodes << "\t";
		nInternalOut << dtss->numTotalInternalNodes << "\t";
		falseHitsOut.flush();
		timeOut.flush();
		totimeOut.flush();
		iosOut.flush();
		sizeOut.flush();
		nLeavesOut.flush();
		nInternalOut.flush();
	}

	falseHitsOut << endl;
	timeOut << endl;
	totimeOut << endl;
	iosOut << endl;
	sizeOut << endl;
	nLeavesOut << endl;
	nInternalOut << endl;

	falseHitsOut.close();
	timeOut.close();
	totimeOut.close();
	iosOut.close();
	sizeOut.close();
	nLeavesOut.close();
	nInternalOut.close();

}
