#include "../common/pch.h"

void Node::init(int enc_s, int enc, Graph *graph){
	G = graph;
	encoding_size = enc_s;
	encoding = enc;
	covered = 0;
	covering = true;
	TO = -1; // this signifies that the topological order has not been assigned to the node yet
}

void Node::setEdges(){
	int adjacent_node; 
	for(int i=0; i<encoding_size; i++){
		adjacent_node = encoding ^ (1 << i);
		if((encoding | adjacent_node) == encoding)
			out.push_back(adjacent_node);
		else
			in.push_back(adjacent_node);
	}
}

bool Node::canBeDeleted(){
	// Returns true if the node does not have even one child with in-degree 1.
	for(int i=0; i<out.size(); i++)
		if(G->nodes_table[out[i]].in.size()==1)
			return false;
	return true;
}

void Node::eraseEdges(){
	// Erase the edges from/to this deleted node
	int i;
	if(in.size()!=0)
		for(i=0;i<in.size();i++)
			eraseElement(encoding, G->nodes_table[in[i]].out);
	if(out.size()!=0)
		for(i=0;i<out.size();i++)
			eraseElement(encoding, G->nodes_table[out[i]].in);
}

void Node::topologicalOrder(int &counter){
	if(out.size()!=0)
		for(int i=0;i<out.size();i++)
			if(G->nodes_table[out[i]].TO==-1) // if the node has not been visited yet
				G->nodes_table[out[i]].topologicalOrder(counter);

	TO = G->graph_size-counter;
	G->TO_table[TO] = encoding;
	counter++;
}

void Node::assignSTParent(){
	int min_pred = INT_MAX, best_parent = INT_MAX;
	
	if(in.size()==0) 
		return;

	for(int i=0;i<in.size();i++){
		if(G->nodes_table[in[i]].pred.size()<min_pred){
			min_pred = G->nodes_table[in[i]].pred.size();
			best_parent = in[i];
		}
	}

	ST_in.push_back(best_parent); // Update the ST parent
	G->nodes_table[best_parent].ST_out.push_back(encoding); // Update the ST children of the node's parent
}

void Node::updatePredSet(){
	if(in.size()==0) 
		return;
	for(int i=0;i<in.size();i++){
		pred[in[i]] = in[i];
		mergeSets(pred, G->nodes_table[in[i]].pred);
	}
}

void Node::assignTableIntervals(int& counter){
	vector<int> interval;
	interval.push_back(++counter);
	G->invert_IT[counter] = encoding;
	if(ST_out.size()!=0)
		for(int i=0;i<ST_out.size();i++)
			G->nodes_table[ST_out[i]].assignTableIntervals(counter);
	interval.push_back(++counter);
	G->IT[encoding].push_back(interval);
}

void Node::assignCoveredInfo(){
	if(in.size()==0)
		return;
	
	int temp_path;
	for(int i=0;i<in.size();i++){
		temp_path = G->nodes_table[in[i]].covered;
		if(existsInList(in[i], ST_in)==-1) // if in[i] is not a parent in the ST
			temp_path++;
		if(temp_path > covered)
			covered = temp_path;
	}
}

void Node::assignCoveringInfo(){
	if(out.size()==0)
		return;

	for(int i=0;i<out.size();i++){
		if((!G->nodes_table[out[i]].covering) || (existsInList(out[i], ST_out)==-1))
			covering = false;
		if(!covering) break;
	}
}

void Node::assignPODominanceInfo(int graph_size){
	int i, j;
	
	// Initialize zero bitstrings
	for(i=0;i<graph_size;i++){
		PODominates[G->TO_table[i]]=0;
	}

	PODominates[encoding] = 1;

	if(out.size()==0)
		return;

	// This node dominates every out[i] along with the nodes dominated by out[i] (transitivity)
	for(i=0;i<out.size();i++){
		PODominates[out[i]] = 1;
		for(j=0;j<graph_size;j++){
			PODominates[G->TO_table[j]] = (PODominates[G->TO_table[j]] |  G->nodes_table[out[i]].PODominates[G->TO_table[j]]);
		}
	}
}
