#ifndef __GENERAL_DEFINITION
#define __GENERAL_DEFINITION


typedef unsigned int uint;

//----BlockFile, CachedBlockFile, Cache---------------------
#define BFHEAD_LENGTH (sizeof(int)*2)    //file header size

#define TRUE 1
#define FALSE 0

#define MDOM 0
#define DOM 1 

#define SEEK_CUR 1
#define SEEK_SET 0
#define SEEK_END 2

typedef char Block[];
//-------------------All-------------------------------------
#define MAXREAL         1e20
#define FLOATZERO       1e-7
#define MAX_DIMENSION   256

#define DIMENSION 10

#define TRUE 1
#define FALSE 0

//#define min(a, b) (((a) < (b))? (a) : (b)  )
//#define max(a, b) (((a) > (b))? (a) : (b)  )

#define MAXSKY 10000000 // by Greg
//-------------------Class and other data types--------------
class BlockFile;  //for BlockFile definition
class Cache;
class Cacheable   //inherit this class if you wish to design an external
                  //memory structure that can be cached
{
public:
	BlockFile *file;
	Cache *cache;
};
 
class CmdIntrpr{  //this is the class of command interpretor.  for a new rtree decescendent
                  //inherit this class to obtain tailored command interpretor
public:
	int cnfrm_cmd(char *_msg){
		char c = ' ';
		while (c != 'N' && c != 'Y'){
			printf("%s (y/n)?", _msg);
			c = getchar(); 
			char tmp;
			while ((tmp = getchar()) != '\n');
			c = toupper(c); 
		}
		if (c == 'N') return 0; else return 1; 
	}
  
	void get_cmd(char *_msg, char *_cmd){
		printf("%s", _msg);  
		char *c = _cmd;
		while (((*c) = getchar()) != '\n')
			c++;
		*c = '\0'; 
	} 

	virtual bool build_tree(char *_tree_fname, char *_data_fname, int _b_len, int _dim, int _csize) = 0;
	virtual void free_tree() = 0;
	virtual int qry_sngle(float *_mbr, int *_io_count) = 0;
	virtual void run() = 0;
	virtual void version() = 0;
};
  
enum SECTION {OVERLAP, INSIDE, S_NONE};
enum R_OVERFLOW {SPLIT, REINSERT, NONE};
enum R_DELETE {NOTFOUND,NORMAL,ERASED};
typedef float *floatptr;
  
struct SortMbr{
    int dimension;
    float *mbr;
    float *center;
    int index;
};

struct SortTuple{
	int id;
	int dimension;
	float score;
    float *bounces;
};

struct BranchList{
    int entry_number;
    float mindist;
    float minmaxdist;
    bool section;
};


void error(char *_errmsg, bool _terminate);
float area(int dimension, float *mbr);
float margin(int dimension, float *mbr);
float overlap(int dimension, float *r1, float *r2);
float* overlapRect(int dimension, float *r1, float *r2);
float* overlapRect_dynamic(int dimension, float *r1, float *r2, bool *func); // by Greg
float objectDIST(float *p1, float *p2);
float MINMAXDIST(float *Point, float *bounces);
float MINDIST(float *Point, float *bounces);
float MAXDIST(float *p, float *bounces, int dim);
float MbrMINDIST(float *_m1, float *_m2, int _dim);
float MbrMAXDIST(float *_m1, float *_m2, int _dim);
float mindist(float *p, float *bounces, int dimension, float *bound); // by Greg
float mindist_dynamic(float *p, float *bounces, int dimension, float *bound, bool *func); // by Greg
float minmaxdist(float *Point, float *bounces, int dimension, float *bound); // by Greg
float minmaxdist_dynamic(float *p, float *bounces, int dimension, float *bound, bool *func); // by Greg
bool inside(float &p, float &lb, float &ub);
void enlarge(int dimension, float **mbr, float *r1, float *r2);
bool inside(float *v, float *mbr, int dimension);
bool mbr_inside(float *mbr1, float *mbr2, int dimension); // by Greg
bool mbr_inside_dynamic(float *mbr1, float *mbr2, int dimension, bool *func); // by Greg
bool point_inside_mbr(float *p, float *mbr2, int dimension); // by Greg
bool boundedFound(float *bounces, float *bound, int dimension); // by Greg
bool boundedFound_dynamic(float *bounces, float *bound, int dimension, bool *func); // by Greg
bool cornerFound(float *bounces, float *bound, int dimension); //by Greg
void printTime(long elapsed); // by Greg
int pruneBrunchList(float *nearest_distanz, const void *activebrunchList, int n);
int boundedpruneBrunchList(float *nearest_distanz, const void *activebrunchList, int n); // by Greg
float* overlapRectPt(int dimension, float *r1, float *r2); // by Greg
bool section(int dimension, float *mbr1, float *mbr2);
bool section_dynamic(int dimension, float *mbr1, float *mbr2, bool *func); // by Greg
bool section_c(int dimension, float *mbr1, float *center, float radius);
int sort_lower_mbr(const void *d1, const void *d2);
int sort_upper_mbr(const void *d1, const void *d2);
int sort_center_mbr(const void *d1, const void *d2);
int sortmindist(const void *element1, const void *element2);
int sort_tuple_score(const void *d1, const void *d2); // by Greg

#ifdef UNIX
void strupr(char *_msg);
#endif

#endif
