#include "../common/pch.h"


RTree::RTree(const char *fname, int _b_length, Cache *c, int _dimension){
  //use this constructor to build a new tree
    file = new BlockFile(fname, _b_length);
    cache = c;

    re_data_cands = new LinList();
	deletelist = new LinList();

    dimension = _dimension;
    root = 0;
    root_ptr = NULL;
    root_is_data = TRUE;
    num_of_data = num_of_inodes = num_of_dnodes = 0;

    root_ptr = new RTNode(this);
	  //note that when a tree is constructed, the root is automatically created
	  //though at this time there is no entry in the root yet.
    num_of_dnodes++;
    root_ptr -> level = 0;
    root = root_ptr -> block;
	delete root_ptr;
	root_ptr = NULL;
}

RTree::RTree(const char *fname, Cache *c)
  //use this constructor to restore a tree from a file
{
    file = new BlockFile(fname, 0);
    cache =c;

    re_data_cands = new LinList();
	deletelist = new LinList();

    char *header = new char [file->get_blocklength()];
    file -> read_header(header);
    read_header(header);
	delete [] header;

    root_ptr = NULL;
}

RTree::RTree(const char *inpname, const char *fname, int _b_length, Cache *c, int _dimension){
  // construct new R-tree from a specified input textfile with rectangles
    Entry *d;
    FILE *fp;
    file = new BlockFile(fname, _b_length);
    cache =c;

    re_data_cands = new LinList();
	deletelist = new LinList();

    char *header = new char [file->get_blocklength()];
    read_header(header);
	delete [] header;

    dimension = _dimension;
    root = 0;
    root_ptr = NULL;
    root_is_data = TRUE;
    num_of_data = num_of_inodes = num_of_dnodes = 0;

    root_ptr = new RTNode(this);
    num_of_dnodes++;
    root_ptr -> level = 0;
    root = root_ptr->block;

	int record_count = 0;

    if((fp = fopen(inpname,"r")) == NULL){
      delete this;
      error("Cannot open R-Tree text file", TRUE);
    }
    else{
	  int id=0;
      while (!feof(fp)){
		record_count ++;
		id ++; //disable this variable if the id comes with the dataset

		if (record_count%1000 == 0){
			for (int i = 0; i < 79; i ++)  //clear a line
			  printf("\b");
			printf("inserting object %d", record_count);
		}

		//fscanf(fp, "%d ", &id); // disable if the id does not come with the dataset
		d = new Entry(dimension, NULL);
    	d -> son = id;
		for (int i=0; i<dimension; i++){
    		fscanf(fp, "%f", &(d->bounces[2*i]));
			d->bounces[2*i+1] = d->bounces[2*i];
		}
		fscanf(fp, "\n");

    	insert(d);
      }
    }

	fclose(fp);

	printf("\n");
	delete root_ptr;
	root_ptr = NULL;
}

RTree::~RTree(){
	char *header = new char[file -> get_blocklength()];
    write_header(header);
    file->set_header(header);
    delete [] header;

    if (root_ptr != NULL){
        delete root_ptr;
        root_ptr = NULL;
    }

	if (cache)
      cache -> flush();

    delete file;
    delete re_data_cands;
	delete deletelist;

  //printf("This R-Tree contains %d internal, %d data nodes and %d data\n",
	//	num_of_inodes, num_of_dnodes, num_of_data);
}

bool RTree::delete_entry(Entry *d){
	load_root();

	R_DELETE del_ret;
	del_ret=root_ptr->delete_entry(d);

	if (del_ret == NOTFOUND) return false;
	if (del_ret == ERASED)
		error("RTree::delete_entry--The root has been deleted\n",true);

	if (root_ptr -> level > 0 && root_ptr -> num_entries == 1){
		//there is only one entry in the root but the root
		//is not leaf.  in this case, the child of the root is exhalted to root
		root = root_ptr -> entries[0].son;
		delete root_ptr;
		root_ptr = NULL;
		load_root();
		num_of_inodes--;
	}

	//Now will reinsert the entries
	while (deletelist -> get_num() > 0){
		Linkable *e;
		e = deletelist -> get_first();
		Entry *new_e = new Entry(dimension, NULL);
		new_e -> set_from_Linkable(e);
		deletelist -> erase();
		insert(new_e);
	}

	delete root_ptr;
	root_ptr = NULL;

	return true;
}

bool RTree::FindLeaf(Entry *e){
	load_root();
	return root_ptr -> FindLeaf(e);
}

void RTree::insert(Entry* d){
    int i, j;
    RTNode *sn;
    RTNode *nroot_ptr;
    int nroot;
    Entry *de;
    R_OVERFLOW split_root;
    Entry *dc;
    float *nmbr;

    // load root into memory
    load_root();

    // no overflow occured until now
    re_level = new bool[root_ptr -> level + 1];
    for (i = 0; i <= root_ptr -> level; i++)
        re_level[i] = FALSE;

    // insert d into re_data_cands as the first entry to insert
    // make a copy of d because it should be erased later
    Linkable *new_link;
	new_link = d -> gen_Linkable();
	re_data_cands -> insert(new_link);

	delete d;  //we follow the convention that the entry will be deleted when insertion finishes

    j = -1;
    while (re_data_cands -> get_num() > 0){
        // first try to insert data, then directory entries
	    Linkable *d_cand;
		d_cand = re_data_cands -> get_first();
        if (d_cand != NULL){
            // since "erase" deletes the data itself from the
            // list, we should make a copy of the data before
            // erasing it
			dc = new Entry(dimension, NULL);
            dc -> set_from_Linkable(d_cand);
            re_data_cands -> erase();

            // start recursive insert with root
			split_root = root_ptr -> insert(dc, &sn);
        }
        else
	        error("RTree::insert: inconsistent list re_data_cands", TRUE);

    	if (split_root == SPLIT){
    	// insert has lead to split --> new root-page with two sons (i.e. root and sn)
    	    nroot_ptr = new RTNode(this);
    	    nroot_ptr -> level = root_ptr -> level + 1;
    	    num_of_inodes++;
    	    nroot = nroot_ptr -> block;

    	    de = new Entry(dimension, this);
    	    nmbr = root_ptr -> get_mbr();
    	    memcpy(de->bounces, nmbr, 2*dimension*sizeof(float));
    	    delete [] nmbr;
    	    de->son = root_ptr->block;
    	    de->son_ptr = root_ptr;
    	    nroot_ptr -> enter(de);

    	    de = new Entry(dimension, this);
    	    nmbr = sn -> get_mbr();
    	    memcpy(de -> bounces, nmbr, 2*dimension*sizeof(float));
    	    delete [] nmbr;
    	    de -> son = sn -> block;
    	    de -> son_ptr = sn;
    	    nroot_ptr->enter(de);

    	    root = nroot;
            root_ptr = nroot_ptr;

            root_is_data = FALSE;
        }
        j++;
    }

    num_of_data++;

    delete [] re_level;
	delete root_ptr;
	root_ptr = NULL;
}

void RTree::load_root(){
    if (root_ptr == NULL)
        root_ptr = new RTNode(this, root);
}

void RTree::NNQuery(float *QueryPoint, SortedLinList *res){
      float nearest_distanz;

      // load root node into main memory
      load_root();
      nearest_distanz = MAXREAL;
      root_ptr->NNSearch(QueryPoint,res,&nearest_distanz);

	  delete root_ptr;
	  root_ptr = NULL;
}

int RTree::rangeQuery(float *mbr){
    load_root();
    int cnt=root_ptr -> rangeQuery(mbr);
	delete root_ptr;
	root_ptr = NULL;
	return cnt;
}

void RTree::read_header(char *buffer){
    int i;

    memcpy(&dimension, buffer, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&num_of_data, &buffer[i], sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&num_of_dnodes, &buffer[i], sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&num_of_inodes, &buffer[i], sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&root_is_data, &buffer[i], sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&root, &buffer[i], sizeof(root));
    i += sizeof(root);
}

void RTree::write_header(char *buffer){
    int i;

    memcpy(buffer, &dimension, sizeof(dimension));
    i = sizeof(dimension);

    memcpy(&buffer[i], &num_of_data, sizeof(num_of_data));
    i += sizeof(num_of_data);

    memcpy(&buffer[i], &num_of_dnodes, sizeof(num_of_dnodes));
    i += sizeof(num_of_dnodes);

    memcpy(&buffer[i], &num_of_inodes, sizeof(num_of_inodes));
    i += sizeof(num_of_inodes);

    memcpy(&buffer[i], &root_is_data, sizeof(root_is_data));
    i += sizeof(root_is_data);

    memcpy(&buffer[i], &root, sizeof(root));
    i += sizeof(root);
}

void RTree::BBS_subspace(XxkHeap *_hp, list<float> &result, bool *_active_dim){
	int i, j;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greather than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j++)
		he->bounces[j] = 0.0;
	_hp->push( he );

	float *b = new float[dimension];
	float *bounces = new float[dimension];

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;

		for (i = 0; i < dimension; i ++){
			if (_active_dim[i])
				bounces[i] = he->bounces[2 * i];
			else
				bounces[i] = 0;
		}

		if (level==0){
			if (!sky_dom_iter(dimension, result, bounces)){
				for(j=0;j<dimension;j++){
					result.push_back(bounces[j]);
				}
			}
		}
		else{
			if (!sky_dom_iter(dimension, result, bounces)){
				RTNode *child = new RTNode(this, son);
				//TSS_D_accesses++;

				for (i = 0; i < child->num_entries; i ++){
					for (j = 0; j < dimension; j ++) {
						if (_active_dim[j])
							b[j] = child->entries[i].bounces[2 * j];
						else
							b[j] = 0;
					}

					if (!sky_dom_iter(dimension, result, b)){
						HeapEntry *tmp_he = new HeapEntry();
						tmp_he->init_HeapEntry( dimension );
						tmp_he->son1 = child->entries[i].son;
						tmp_he->level = child->level;
						tmp_he->key = 0;

						for (j = 0; j < dimension; j ++)
							if (_active_dim[j])
								tmp_he->key += b[j]; // Manhattan distance

						memcpy( tmp_he->bounces, child->entries[i].bounces, 2 * dimension * sizeof( float ) );
						_hp->push( tmp_he );
					}
				}
				delete child;
			}
		}
		delete he;
	}
	delete [] bounces;
	delete [] b;

	return;
}

/*
void RTree::TSS_D_skyline(Graph **G, int TO_dim, int PO_dim, XxkHeap *_hp, list<float> &result, bool *_active_dim, float* PO_values, int maxValue){
	int i, j;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greather than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j++)
		he->bounces[j] = 0.0;
	_hp->push( he );

	float *b = new float[TO_dim+2*PO_dim];
	float *bounces = new float[TO_dim+2*PO_dim];

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;

		for (i = 0; i < TO_dim; i ++){
			if (_active_dim[i])
				bounces[i] = he->bounces[2*i];
			else
				bounces[i] = 0;
		}
		for(i=0;i<PO_dim;i++){
			bounces[TO_dim+2*i] = PO_values[i];
			bounces[TO_dim+2*i+1] = PO_values[i];
		}

		if (level==0){
			if (!TSS_D_sky_dom_iter(G, TO_dim, PO_dim, result, bounces, maxValue)){
				for(j=0;j<TO_dim;j++){
					//cout<<bounces[j]<<" ";
					result.push_back(bounces[j]);
				}
				for(j=0;j<PO_dim;j++){
					//cout<<bounces[TO_dim+2*j]<<" ";
					result.push_back(bounces[TO_dim+2*j]);
				}
				//cout<<endl;
			}
		}
		else{
			if (!TSS_D_sky_dom_iter(G, TO_dim, PO_dim, result, bounces, maxValue)){
				RTNode *child = new RTNode(this, son);
//				TSS_D_accesses++;

				for (i = 0; i < child->num_entries; i ++){
					for (j = 0; j < TO_dim; j ++) {
						if (_active_dim[j])
							b[j] = child->entries[i].bounces[2 * j];
						else
							b[j] = 0;
					}
					for(j=0;j<PO_dim;j++){
						b[TO_dim+2*j] = PO_values[j];
						b[TO_dim+2*j+1] = PO_values[j];
					}
					if (!TSS_D_sky_dom_iter(G, TO_dim, PO_dim, result, b, maxValue)){
						HeapEntry *tmp_he = new HeapEntry();
						tmp_he->init_HeapEntry( dimension );
						tmp_he->son1 = child->entries[i].son;
						tmp_he->level = child->level;
						tmp_he->key = 0;

						for (j = 0; j < dimension; j ++)
							if (_active_dim[j])
								tmp_he->key += b[j]; // Manhattan distance

						memcpy( tmp_he->bounces, child->entries[i].bounces, 2 * dimension * sizeof( float ) );
						_hp->push( tmp_he );
					}
				}
				delete child;
			}
		}
		delete he;
	}
	delete [] bounces;
	delete [] b;

	return;
}
*/

/*
void RTree::SDC_skyline(Graph** G, int TO_dim, int PO_dim, XxkHeap *_hp, list<float> &result, list<float> &global_result, bool *_active_dim){
	int i, j;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greather than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j++)
		he->bounces[j] = 0.0;
	_hp->push( he );

	float *b = new float[dimension];
	float *bounces = new float[dimension];

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;

		for (i = 0; i < dimension; i ++){
			if (_active_dim==NULL)
				bounces[i] = he->bounces[2 * i];
			else
				bounces[i] = 0;
		}

		if (level==0){
			if (!SDC_sky_dom_iter(G, TO_dim, PO_dim, dimension, result, bounces, DOM)){
				bool retu = false;
				/// check for dominance with globalResult


//				int _dim = TO_dim + 2*PO_dim;
//				list<float>::iterator giter = global_result.begin();
//				while (giter!=global_result.end()){
//					float *gp = new float[_dim];
//					for (j=0; j<_dim; j++){
//						gp[j] = *giter;
//						giter++;
//					}
//
//					if(SDC_dominate(G, TO_dim, PO_dim, _dim, gp, bounces)){
//						retu = true;
//						SDC_false_hits++;
//						delete gp;
//						break;
//					}
//					delete gp;
//				}


				if(!retu){ // the point is not dominated
					for(j=0;j<dimension;j++){
//					cout << bounces[j] << " ";
						result.push_back(bounces[j]);
					}

				}
			}


//				cout << endl;
//				cout << "result size " << result.size() << endl;
		}
		else{
			if (!SDC_sky_dom_iter(G, TO_dim, PO_dim, dimension, result, bounces, MDOM)){
				RTNode *child = new RTNode(this, son);
				SDC_accesses++;

				for (i = 0; i < child->num_entries; i ++){
					for (j = 0; j < dimension; j ++) {
						if (_active_dim==NULL)
							b[j] = child->entries[i].bounces[2 * j];
						else
							b[j] = 0;
					}

					if (!SDC_sky_dom_iter(G, TO_dim, PO_dim, dimension, result, b, MDOM)){
						HeapEntry *tmp_he = new HeapEntry();
						tmp_he->init_HeapEntry( dimension );
						tmp_he->son1 = child->entries[i].son;
						tmp_he->level = child->level;
						tmp_he->key = 0;

						for (j = 0; j < dimension; j ++)
							if (_active_dim == NULL)
								tmp_he->key += b[j]; // Manhattan distance

						memcpy( tmp_he->bounces, child->entries[i].bounces, 2 * dimension * sizeof( float ) );
						_hp->push( tmp_he );
					}
				}
				delete child;
			}
		}
		delete he;
	}
	delete [] bounces;
	delete [] b;

	return;
}
*/


/*
bool RTree::SDC_sky_dom_iter(Graph** G, int TO_dim, int PO_dim, int _dim, list<float> &result, float *_pt, int dom){
	bool ret = false;

	list<float>::iterator iter, fiter;

	iter = result.begin();
	while (iter!=result.end()){
		fiter = iter;
		float *s_pt = new float[_dim];
		for (int j=0; j<_dim; j++){
			s_pt[j] = *fiter;
			fiter++;
		}

		bool resultPointDeleted = false;

		if(dom == DOM){
			if(SDC_dominate(G, TO_dim, PO_dim, _dim, s_pt, _pt)){
				ret = true;
			}
			else if(SDC_dominate(G, TO_dim, PO_dim, _dim, _pt, s_pt)){
				SDC_false_hits++;
				for (uint jj=0; jj<_dim; jj++){
					iter = result.erase(iter);
				}
				resultPointDeleted = true;
			}
		}


		if(dom == MDOM) {
			ret = SDC_m_dominate(_dim, s_pt, _pt);
		}

		delete[] s_pt;

		if(ret){
			break;
		}

		if (!resultPointDeleted){
			for (uint jj=0; jj<_dim; jj++){
				iter++;
			}
		}
	}
	return ret;
}
*/

/*
bool RTree::TSS_S_sky_dom_iter(Graph **G, int TO_dim, int PO_dim, vector<float> &result, float* _pt, int maxValue){
	int i, j;
	bool ret = false;
	vector<float>::iterator iter;

	iter = result.begin();

	// Extract TO_point
	float *TO_point = new float[TO_dim];
	for(i=0;i<TO_dim;i++)
		TO_point[i] = _pt[2*i];

	while (iter!=result.end()){
		// Extract a result point
		float *s_pt = new float[TO_dim+PO_dim];
		for (int j=0; j<(TO_dim+PO_dim); j++){
			s_pt[j] = *iter;
			iter++;
		}

		// Check dominance on the total dimensions
		if (!dominate(TO_dim, s_pt, TO_point) || !TSS_PO_dominate(G, PO_dim, s_pt+TO_dim, _pt+2*TO_dim, maxValue)){
			delete[] s_pt;
			continue;
		}

		delete[] s_pt;
		ret = true;
		break;
	}

	delete[] TO_point;
	return ret;
}
*/

/*
bool RTree::TSS_D_sky_dom_iter(Graph** G, int TO_dim, int PO_dim, list<float> &result, float* _pt, int maxValue){
	int i, j;
	bool ret = false;
	list<float>::iterator iter;

	iter = result.begin();

	// Extract TO_point
	float *TO_point = new float[TO_dim];
	for(i=0;i<TO_dim;i++)
		TO_point[i] = _pt[i];

	char c;
	while (iter!=result.end()){
		// Extract a result point
		float *s_pt = new float[TO_dim+PO_dim];
		for (int j=0; j<(TO_dim+PO_dim); j++){
			s_pt[j] = *iter;
			iter++;
		}

		// Check dominance on the total dimensions

		if (!TO_dominate(TO_dim, s_pt, TO_point) || !TSS_PO_dominate(G, PO_dim, s_pt+TO_dim, _pt+TO_dim, maxValue)){
			delete[] s_pt;
			continue;
		}

		delete[] s_pt;
		ret = true;
		break;
	}

	delete[] TO_point;

	return ret;
}
*/


/// deprecated use _iter
/*
bool RTree::SDC_sky_dom(Graph** G, int TO_dim, int PO_dim, int _dim, list<float> &result, float *_pt, int dom){
	bool ret = false;
	for (int i = 0; i < result.size()/_dim; i ++){
		float *temp_float = new float[_dim];
		for(int j=0;j<_dim;j++)
			temp_float[j] = result[i*_dim +j];
		float *s_pt = temp_float;

		if(dom == DOM){
			if(SDC_dominate(G, TO_dim, PO_dim, _dim, s_pt, _pt)){
				ret = true;
			}
			else if(SDC_dominate(G, TO_dim, PO_dim, _dim, _pt, s_pt)){
				cout << "DANGER" << endl;
				result.erase(result.begin()+i*_dim, result.begin() + i*_dim + _dim);  // delete the result
			}
		}

		if(dom == MDOM)
			ret = SDC_m_dominate(_dim, s_pt, _pt);

		delete[] s_pt;

		if(ret)
			break;
	}
	return ret;
}
*/

// BBS for computing the skyband in a subspace

// _hp: heap to be used
// _rslt: result skyline list to be used
// _active_dim: a boolean array whose size equals the dimensionality of the R-tree
//              the i-th element equals 'true' if the i-th dimension is involved in the subspace

/*
void RTree::BBS_subspace_band( XxkHeap *_hp, float *_rslt, int &_rsltcnt, bool *_active_dim, int nTopK ){
	int i, j;
	_rsltcnt = 0;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greather than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j ++)
		he->bounces[j] = 0.0;

	_hp->push( he );

	float *bounces = new float[dimension];
	float *b = new float[dimension];

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;
		for (int i = 0; i < dimension; i ++){
			if (_active_dim[i])
				bounces[i] = he->bounces[2 * i];
			else
				bounces[i] = 0;
		}

		if (level==0){
			if (!skyband_dom(dimension, _rslt, _rsltcnt, bounces, nTopK)){
				memcpy(&_rslt[dimension * _rsltcnt], bounces, sizeof(float) * dimension);
				_rsltcnt ++;
			}
		}
		else{
			if (!skyband_dom(dimension, _rslt, _rsltcnt, bounces, nTopK)){
				RTNode *child = new RTNode(this, son);

				for (i = 0; i < child->num_entries; i ++){
					for (j = 0; j < dimension; j ++) {
						if (_active_dim[j])
							b[j] = child->entries[i].bounces[2 * j];
						else
							b[j] = 0;
					}

					if (!skyband_dom(dimension, _rslt, _rsltcnt, b, nTopK)){
						HeapEntry *tmp_he = new HeapEntry();
						tmp_he->init_HeapEntry( dimension );

						tmp_he->son1 = child->entries[i].son;
						tmp_he->level = child->level;

						tmp_he->key = 0;

						for (j = 0; j < dimension; j ++){
							if (_active_dim[j]){
								tmp_he->key = b[j];
								break;
							}
						}

						memcpy( tmp_he->bounces, child->entries[i].bounces, 2 * dimension * sizeof( float ) );

						_hp->push( tmp_he );
					}
				}
				delete child;
			}
		}

		delete he;
	}
	delete [] b;
	delete [] bounces;

	return;
}


// BBS for computing the topK results in a subspace
// _hp: heap to be used
// _rslt: result skyline list to be used
// _dim_weight: an array whose size equals the dimensionality of the R-tree
//              the i-th element equals the weight on the the i-th dimension

void RTree::BBS_top_k( XxkHeap *_hp, float *_rslt, int &_rsltcnt, float *_dim_weight, int nTopK ){
	int i, j;
	_rsltcnt = 0;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greater than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j++)
		he->bounces[j] = 0.0;

	_hp->push( he );

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;

		if (level==0){
			memcpy(&_rslt[dimension * _rsltcnt], he->bounces, sizeof(float) * dimension);
			++_rsltcnt;
			delete he;
			if (_rsltcnt == nTopK){
				while(_hp->empty() == false){
					he = _hp->top();
					_hp->pop();
					delete he;
				}
				break;
			}
		}
		else{
			RTNode *child = new RTNode(this, son);
			for (i = 0; i < child->num_entries; i ++){
				HeapEntry *tmp_he = new HeapEntry();
				tmp_he->init_HeapEntry( dimension );

				tmp_he->son1 = child->entries[i].son;
				tmp_he->level = child->level;

				tmp_he->key = 0;
				for (j = 0; j < dimension; j ++){
					tmp_he->bounces[j] = child->entries[i].bounces[2 * j];
					tmp_he->key += _dim_weight[j] * tmp_he->bounces[j];
				}

				_hp->push( tmp_he );
			}

			delete he;
			delete child;
		}
	}

	return;
}
*/

//this function checks whether the new point is dominated by any
//skyline point already found
//dim:
//rslt: an array with size dim * rsltcnt. the first dim numbers capture
//      the coordinates of the first point, the next dim numbers for the
//      the next point, and so on.
//rsltcnt: number of skyline points so far
//pt: the point being tested


/*
void RTree::TSS_S_skyline(Graph **G, int TO_dim, int PO_dim, XxkHeap *_hp, vector<float> &result, int maxValue){
	int i, j;

	HeapEntry *he=new HeapEntry();
	he->dim = dimension;
	he->son1 = root;
	he->key = 0;
	he->level = 1; //this is not important but make sure it is greather than 0
	he->bounces = new float[2 * dimension];
	for (j = 0; j < 2 * dimension; j++)
		he->bounces[j] = 0.0;
	_hp->push( he );

	float *b = new float[2*dimension];
	float *bounces = new float[2*dimension];

	while (_hp->empty() == false){
		he = _hp->top();
		_hp->pop();

		int son = he->son1;
		int level = he->level;

		for (i = 0; i < 2*dimension; i ++){
			bounces[i] = he->bounces[i];
		}

		if (level==0){
			if (!TSS_S_sky_dom_iter(G, TO_dim, PO_dim, result, bounces, maxValue)){
//				progrTSS << TSS_accesses << endl;
				for(j=0;j<dimension;j++){
					result.push_back(bounces[2*j]);
					//cout<<bounces[2*j]<<" ";
				}
				//cout<<endl;
			}
		}
		else{
			if (!TSS_S_sky_dom_iter(G, TO_dim, PO_dim, result, bounces, maxValue)){
				RTNode *child = new RTNode(this, son);
				TSS_accesses++;

				for (i = 0; i < child->num_entries; i ++){
					for (j = 0; j < 2*dimension; j ++)
							b[j] = child->entries[i].bounces[j];

					if (!TSS_S_sky_dom_iter(G, TO_dim, PO_dim, result, b, maxValue)){
						HeapEntry *tmp_he = new HeapEntry();
						tmp_he->init_HeapEntry( dimension );
						tmp_he->son1 = child->entries[i].son;
						tmp_he->level = child->level;
						tmp_he->key = 0;

						for (j = 0; j < dimension; j ++)
							tmp_he->key += b[2*j]; // Manhattan distance

						memcpy( tmp_he->bounces, child->entries[i].bounces, 2 * dimension * sizeof( float ) );
						_hp->push( tmp_he );
					}
				}
				delete child;
			}
		}
		delete he;
	}

	delete [] bounces;
	delete [] b;

	return;
}
*/

bool RTree::sky_dom_iter(int _dim, list<float> &result, float *_pt){
	bool ret = false;

	list<float>::iterator iter;

	iter = result.begin();
	while (iter!=result.end()){
		float *s_pt = new float[_dim];
		for (int j=0; j<_dim; j++){
			s_pt[j] = *iter;
			iter++;
		}

		if (TO_dominate(_dim, s_pt, _pt)){
			ret = true;
			delete[] s_pt;
			break;
		}
	}

	return ret;
}

/// deprecated use _iter
/*
bool RTree::sky_dom(int _dim, vector<float> &result, float *_pt){
	bool ret = false;
	for (int i = 0; i < result.size()/_dim; i ++){
		float *temp_float = new float[_dim];
		for(int j=0;j<_dim;j++){
			temp_float[j] = result[i*_dim +j];
		}
		float *s_pt = temp_float;
		if (dominate(_dim, s_pt, _pt)){
			ret = true;
			delete[] s_pt;
			break;
		}
		delete[] s_pt;
	}
	return ret;
}
*/

/*
bool RTree::skyband_dom( int _dim, float *_rslt, int _rsltcnt, float *_pt, int nTopK ){
	bool ret = false;
	int counter = 0;
	for (int i = 0; i < _rsltcnt; i ++){
		float *s_pt = &_rslt[i * _dim];
		if (dominate(_dim, s_pt, _pt)){
			++counter;
			if (counter >= nTopK){
				ret = true;
				break;
			}
		}
	}
	return ret;
}
*/

//bulkloads the tree by sorting the data points according to the hilbert values
//fname: the name of the dataset file
//dim: the dimensionality
//size: the dataset size
//number of pits to compute the hilbert values

int hil_compare( const void *_pt1, const void *_pt2 ){
	hilmbr_ptr pt1 = (hilmbr_ptr) _pt1;
	hilmbr_ptr pt2 = (hilmbr_ptr) _pt2;

	if (pt1->hil_v < pt2->hil_v) return -1;
	else if (pt1->hil_v > pt2->hil_v) return 1;
	else return 0;
}


void RTree::bulkloadDUMMY(FILE *fp, const char *_fname, int _dim, int _size, int _num_bit){
	int i;
	hilmbr *mbr = NULL;

//	ifstream fin;
//	fin.open(_fname, ios::in);
//	if (fin.fail()){
//		cerr << "fin.open() failed on " << _fname << endl;
//	}
//	fin.close();

//	FILE *fp = fopen(_fname, "r");
	string msg = _fname;
	msg += "\n";
	msg = "dummy could not open data source file " + msg;
	if (!fp)
		error((char*) msg.c_str(), true);
//	fclose(fp);

	//printf("initilizing the memory needed...\n");
		mbr = new hilmbr[_size];
		for (i = 0; i < _size; i ++)
			mbr[i].bounces = new float[2 * _dim];

		//printf("reading the data set...\n");
		int cnt = 0;
		while (!feof(fp) && cnt < _size){
			mbr[cnt].block = cnt;
			for (i = 0; i < _dim; i ++){
				fscanf(fp, " %f", &(mbr[cnt].bounces[2 * i]));
				mbr[cnt].bounces[2 * i + 1] = mbr[cnt].bounces[2 * i];
			}
			fscanf(fp, "\n"); // dsachar

			//caculate the hilbert value of this point
			mbr[cnt].hil_v = cal_hilbert_mbr(5, _dim, mbr[cnt].bounces, 10000);
			if (mbr[cnt].hil_v < 0)
				error("error: negative hilbert values obtained.\n", true);

			cnt ++;
		}

//		fclose(fp); // dsachar
		if (cnt < _size){
			printf("error: the input dataset is smaller than the specified size\n");
			if (mbr){
				for (i = 0; i < _size; i ++)
					delete[] mbr[i].bounces;
			}
			delete []mbr;
	//		fclose(fp); // dsachar
			return;
		}
	//	else if (!feof(fp)){ // dsachar
	////	else if (feof(fp)){ // dsachar
	//		printf("warning: the input dataset is larger than the specified size\n");
	//		if (mbr){
	//			for (i = 0; i < _size; i ++)
	//				delete []mbr[i].bounces;
	//		}
	//		delete []mbr;
	//		fclose(fp); // dsachar
	//		return;
	//	}
	//	fclose(fp); // dsachar

		//printf("sorting the dataset...\n");
		qsort(mbr, _size, sizeof(hilmbr), hil_compare);

		int level = 0;
		int mbr_num = _size;

		while (level == 0 || mbr_num > 1){
			//printf("building nodes at level %d...\n", level);
			int new_mbr_num = 0;

			RTNode *nd = new RTNode(this); nd->level = level;

			if (level == 0)
				num_of_dnodes ++;
			else
				num_of_inodes ++;

			for (i = 0; i < mbr_num; i ++){
				memcpy(nd->entries[nd->num_entries].bounces, mbr[i].bounces, 2 * sizeof(float) * dimension);
				nd->entries[nd->num_entries].son = mbr[i].block;

				nd->num_entries ++;
				if (nd->num_entries == nd->capacity){
					float *ndmbr = nd->get_mbr();
					memcpy(mbr[new_mbr_num].bounces, ndmbr, 2 * sizeof(float) * dimension);
					mbr[new_mbr_num].block = nd->block;
					mbr[new_mbr_num].hil_v = cal_hilbert_mbr(15, _dim, mbr[new_mbr_num].bounces, 10000);
					new_mbr_num ++;
					delete []ndmbr;
					delete nd;
					nd = new RTNode(this); nd->level = level;
					if (level == 0)
						num_of_dnodes ++;
					else
						num_of_inodes ++;
				}
			}
			float *ndmbr = nd->get_mbr();
			memcpy(mbr[new_mbr_num].bounces, ndmbr, 2 * sizeof(float) * dimension);
			mbr[new_mbr_num].block = nd->block;
			mbr[new_mbr_num].hil_v = cal_hilbert_mbr(15, _dim, mbr[new_mbr_num].bounces, 10000);
			new_mbr_num ++;
			delete []ndmbr;
			delete nd;

			level ++; mbr_num = new_mbr_num;
			qsort(mbr, mbr_num, sizeof(hilmbr), hil_compare);
		}

		root = mbr[0].block; num_of_data = _size;

		//printf("freeing up the memory...\n");
		if (mbr){
			for (i = 0; i < _size; i ++)
				delete []mbr[i].bounces;
		}
		delete []mbr;
	//	fclose(fp); // dsachar
}


void RTree::bulkload(const char *_fname, int _dim, int _size, int _num_bit){
	int i;
	hilmbr *mbr = NULL;

	FILE *fp = fopen(_fname, "r");
	string msg = _fname;
	msg += "\n";
	msg = "could not open data source file " + msg;
	if (!fp)
		error((char*) msg.c_str(), true);

	//printf("initilizing the memory needed...\n");
	mbr = new hilmbr[_size];
	for (i = 0; i < _size; i ++)
		mbr[i].bounces = new float[2 * _dim];

	//printf("reading the data set...\n");
	int cnt = 0;
	while (!feof(fp) && cnt < _size){
		mbr[cnt].block = cnt;
		for (i = 0; i < _dim; i ++){
			fscanf(fp, " %f", &(mbr[cnt].bounces[2 * i]));
			mbr[cnt].bounces[2 * i + 1] = mbr[cnt].bounces[2 * i];
		}
		fscanf(fp, "\n"); // dsachar

		//caculate the hilbert value of this point
		mbr[cnt].hil_v = cal_hilbert_mbr(5, _dim, mbr[cnt].bounces, 10000);
		if (mbr[cnt].hil_v < 0)
			error("error: negative hilbert values obtained.\n", true);

		cnt ++;
	}

	fclose(fp); // dsachar
	if (cnt < _size){
		printf("error: the input dataset is smaller than the specified size\n");
		if (mbr){
			for (i = 0; i < _size; i ++)
				delete[] mbr[i].bounces;
		}
		delete []mbr;
//		fclose(fp); // dsachar
		return;
	}
//	else if (!feof(fp)){ // dsachar
////	else if (feof(fp)){ // dsachar
//		printf("warning: the input dataset is larger than the specified size\n");
//		if (mbr){
//			for (i = 0; i < _size; i ++)
//				delete []mbr[i].bounces;
//		}
//		delete []mbr;
//		fclose(fp); // dsachar
//		return;
//	}
//	fclose(fp); // dsachar

	//printf("sorting the dataset...\n");
	qsort(mbr, _size, sizeof(hilmbr), hil_compare);

	int level = 0;
	int mbr_num = _size;

	while (level == 0 || mbr_num > 1){
		//printf("building nodes at level %d...\n", level);
		int new_mbr_num = 0;

		RTNode *nd = new RTNode(this); nd->level = level;

		if (level == 0)
			num_of_dnodes ++;
		else
			num_of_inodes ++;

		for (i = 0; i < mbr_num; i ++){
			memcpy(nd->entries[nd->num_entries].bounces, mbr[i].bounces, 2 * sizeof(float) * dimension);
			nd->entries[nd->num_entries].son = mbr[i].block;

			nd->num_entries ++;
			if (nd->num_entries == nd->capacity){
				float *ndmbr = nd->get_mbr();
				memcpy(mbr[new_mbr_num].bounces, ndmbr, 2 * sizeof(float) * dimension);
				mbr[new_mbr_num].block = nd->block;
				mbr[new_mbr_num].hil_v = cal_hilbert_mbr(15, _dim, mbr[new_mbr_num].bounces, 10000);
				new_mbr_num ++;
				delete []ndmbr;
				delete nd;
				nd = new RTNode(this); nd->level = level;
				if (level == 0)
					num_of_dnodes ++;
				else
					num_of_inodes ++;
			}
		}
		float *ndmbr = nd->get_mbr();
		memcpy(mbr[new_mbr_num].bounces, ndmbr, 2 * sizeof(float) * dimension);
		mbr[new_mbr_num].block = nd->block;
		mbr[new_mbr_num].hil_v = cal_hilbert_mbr(15, _dim, mbr[new_mbr_num].bounces, 10000);
		new_mbr_num ++;
		delete []ndmbr;
		delete nd;

		level ++; mbr_num = new_mbr_num;
		qsort(mbr, mbr_num, sizeof(hilmbr), hil_compare);
	}

	root = mbr[0].block; num_of_data = _size;

	//printf("freeing up the memory...\n");
	if (mbr){
		for (i = 0; i < _size; i ++)
			delete []mbr[i].bounces;
	}
	delete []mbr;
//	fclose(fp); // dsachar
}

//bulkloads the tree using the str algorithm
//fname: the name of the dataset file
//dim: the dimensionality
//size: the dataset size

void RTree::str_strap(hilmbr *_mbr, int _num_mbr, hilmbr *_new_mbr, int &_new_num_mbr, int _active_dim,
							int _stop_dim, int _num_straps, int _level){
	int i;
	for (i = 0; i < _num_mbr; i ++){
		if (_mbr[i].bounces[0] < 0 || _mbr[i].bounces[0] > 10000)
			error("error here\n", true);
		_mbr[i].hil_v = (_mbr[i].bounces[2 * _active_dim] + _mbr[i].bounces[2 * _active_dim + 1]) / 2;
	}
	qsort(_mbr, _num_mbr, sizeof(hilmbr), hil_compare);

	if (_active_dim < _stop_dim){ //partition the mbrs into _num_straps straps along the _active_dim-th dimension
		int num_mbr_per_strap = (int) floor((float) _num_mbr / _num_straps);
		if (_num_mbr - (_num_straps - 1) * num_mbr_per_strap < 0)
			error("error in str_strap\n", true);

		for (i = 0; i < _num_straps; i ++){
			str_strap(_mbr + i * num_mbr_per_strap, min(num_mbr_per_strap,
				_num_mbr - i * num_mbr_per_strap), _new_mbr, _new_num_mbr, _active_dim + 1,
				_stop_dim, _num_straps, _level);
		}
	}
	else{
		//start creating nodes
		int new_mbr_num = 0;

		RTNode *nd = new RTNode(this); nd->level = _level;
		if (_level == 0)
			num_of_dnodes ++;
		else
			num_of_inodes ++;

		for (i = 0; i < _num_mbr; i ++){
			memcpy(nd->entries[nd->num_entries].bounces, _mbr[i].bounces, 2 * sizeof(float) * dimension);
			nd->entries[nd->num_entries].son = _mbr[i].block;

			nd->num_entries ++;

			if (nd->num_entries == nd->capacity){
				float *ndmbr = nd->get_mbr();
				memcpy(_new_mbr[_new_num_mbr].bounces, ndmbr, 2 * sizeof(float) * dimension);
				_new_mbr[_new_num_mbr].block = nd->block;
				_new_num_mbr ++;
				delete []ndmbr;
				delete nd;

				nd = new RTNode(this); nd->level = _level;
				if (_level == 0)
					num_of_dnodes ++;
				else
					num_of_inodes ++;
			}
		}
		float *ndmbr = nd->get_mbr();
		memcpy(_new_mbr[_new_num_mbr].bounces, ndmbr, 2 * sizeof(float) * dimension);
		_new_mbr[_new_num_mbr].block = nd->block;
		_new_num_mbr ++;
		delete []ndmbr;
		delete nd;
	}
}

void RTree::bulkload_str(const char *_fname, int _dim, int _size){
	int i;
	if (dimension != _dim)
		error("the dimensionality of the tree is not correct\n", true);
	hilmbr *mbr = NULL;

	FILE *fp = fopen(_fname, "r");
	if (!fp)
		error("could not open the data source file\n", true);

	printf("initilizing the memory needed...\n");
	mbr = new hilmbr[_size];
	for (i = 0; i < _size; i ++)
		mbr[i].bounces = new float[2 * _dim];

	printf("reading the data set...\n");
	int cnt = 0;
	while (!feof(fp) && cnt < _size){
		fscanf(fp, "%d", &(mbr[cnt].block));
		for (i = 0; i < _dim; i ++){
			fscanf(fp, " %f", &(mbr[cnt].bounces[2 * i]));
			mbr[cnt].bounces[2 * i + 1] = mbr[cnt].bounces[2 * i];
		}
		fscanf(fp, "\n");

		cnt ++;
	}

	fclose(fp);
	if (cnt < _size){
		printf("error: the input dataset is smaller than the specified size\n");
		if (mbr){
			for (i = 0; i < _size; i ++)
				delete []mbr[i].bounces;
		}
		delete []mbr;
		return;
	}
	else if (feof(fp)){
		printf("error: the input dataset is larger than the specified size\n");
		if (mbr){
			for (i = 0; i < _size; i ++)
				delete []mbr[i].bounces;
		}
		delete []mbr;
		return;
	}

	hilmbr *new_mbr = new hilmbr[_size];
	for (i = 0; i < _size; i ++)
		new_mbr[i].bounces = new float[2 * _dim];

	RTNode *nd = new RTNode(this, root);
	int capacity = nd->capacity;
	printf("the node capacity = %d\n", capacity);
	delete nd;

	int level = 0;
	int mbr_num = _size;

	num_of_dnodes = 0;
	num_of_inodes = 0;

	while (level == 0 || mbr_num > 1){
		printf("building level %d...\n", level);
		int stop_dim = -1;
		int num_straps = -1;

		int num_nodes = (int) ceil((float)mbr_num / capacity);
		if (num_nodes > pow(2.0, (double)_dim)){
			num_straps = (int) pow(num_nodes, (float) 1/ dimension);
			stop_dim = dimension - 1;
		}
		else{
			stop_dim = (int) (log((double)num_nodes) / log(2.0));
			num_straps = 2;
		}

		printf("\t# of nodes = %d, # of straps = %d, stop dim = %d\n", num_nodes, num_straps, stop_dim);

		int new_mbr_num = 0;
		str_strap(mbr, mbr_num, new_mbr, new_mbr_num, 0, stop_dim, num_straps, level);

		mbr_num = new_mbr_num;
		hilmbr *tmp = new_mbr;
		new_mbr = mbr; mbr = tmp; level ++;

	}

	root = mbr[0].block; num_of_data = _size;

	printf("freeing up the memory...\n");
	for (i = 0; i < _size; i ++)
		delete[] new_mbr[i].bounces;
	delete[] new_mbr;

	for (i = 0; i < _size; i ++)
		delete[] mbr[i].bounces;
	delete[] mbr;

	fclose(fp);
}
