#ifndef __RTREE
#define __RTREE

#include <list>

struct hilmbr{
	int block;
	float *bounces;
	bitmask_t hil_v;
};

typedef hilmbr * hilmbr_ptr;

class RTree : public Cacheable{
public:
	int dimension;
	int num_of_data;
    int num_of_dnodes;
    int num_of_inodes;
	int root;
	bool root_is_data;

	RTNode *root_ptr;
    bool *re_level;
    LinList *re_data_cands;
	LinList *deletelist;

	int na[10];

	RTree(const char *fname, int _b_length, Cache* c, int _dimension);
    RTree(const char *fname, Cache* c);
    RTree(const char *inpname, const char *fname, int _blength, Cache* c, int _dimension);
    ~RTree();

	bool delete_entry(Entry *d);
	bool FindLeaf(Entry *e);
    int get_num() { return num_of_data; }
	void insert(Entry *d);
	void load_root();
	void NNQuery(float *QueryPoint, SortedLinList *res);
	int rangeQuery(float *mbr);
	void read_header(char *buffer);
	void write_header(char *buffer);

	void bulkload(const char *_fname, int _dim, int _size, int _num_bit);
	void bulkloadDUMMY(FILE *fp, const char *_fname, int _dim, int _size, int _num_bit);
	void str_strap(hilmbr *_mbr, int _num_mbr, hilmbr *_new_mbr, int &_new_num_mbr,
		int _active_dim, int _stop_dim, int _num_straps, int _level);
	void bulkload_str(const char *_fname, int _dim, int _size);


	bool sky_dom(int _dim, list<float> &result, float *_pt);
	bool sky_dom_iter(int _dim, list<float> &result, float *_pt);

	void BBS_subspace(XxkHeap *_hp, list<float> &result, bool *_active_dim);
	//bool skyband_dom( int _dim, float *_rslt, int _rsltcnt, float *_pt, int nTopK );
	//void BBS_subspace_band(XxkHeap *_hp, float *_rslt, int &_rsltcnt, bool *_active_dim, int nTopK );
	//void BBS_top_k( XxkHeap *_hp, float *_rslt, int &_rsltcnt, float *_dim_weight, int nTopK );



	// DEPRECATED METHODS defined in deprecated.cpp
	// replaced with methods defined in the RTree subclasses (SDC_RTree, TSS_RTree, etc.)

	// SDC
	bool SDC_sky_dom(Graph **G, int TO_dim, int PO_dim, int _dim, list<float> &result, float *_pt, int dom);
	bool SDC_sky_dom_iter(Graph** G, int TO_dim, int PO_dim, int _dim, list<float> &result, float *_pt, int dom); // with iterator
	void SDC_skyline(Graph **G, int TO_dim, int PO_dim, XxkHeap *_hp, list<float> &result, list<float> &globalResult, bool *_active_dim);


	// TSS_S
	bool TSS_S_sky_dom_iter(Graph **G, int TO_dim, int PO_dim, vector<float> &result, float* _pt, int maxValue);
	void TSS_S_skyline(Graph **G, int TO_dim, int PO_dim, XxkHeap *_hp, vector<float> &result, int maxValue);

	//TSS_D
	void TSS_D_skyline(Graph** G, int TO_dim, int PO_dim, XxkHeap *_hp, list<float> &result, bool *_active_dim, float *PO_values, int maxValue);
	bool TSS_D_sky_dom_iter(Graph** G, int TO_dim, int PO_dim, list<float> &result, float* _pt, int maxValue);



};
#endif // __RTREE
